#include <winver.h>
#include "../hydrasdr.h"

#ifdef GCC_WINDRES
VS_VERSION_INFO		VERSIONINFO
#else
VS_VERSION_INFO		VERSIONINFO	MOVEABLE IMPURE LOADONCALL DISCARDABLE
#endif
  FILEVERSION		HYDRASDR_VER_MAJOR,HYDRASDR_VER_MINOR,HYDRASDR_VER_REVISION,0
  PRODUCTVERSION	HYDRASDR_VER_MAJOR,HYDRASDR_VER_MINOR,HYDRASDR_VER_REVISION,0
  FILEFLAGSMASK		VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS		1
#else
  FILEFLAGS		0
#endif
  FILEOS		VOS__WINDOWS32
  FILETYPE		VFT_DLL
  FILESUBTYPE		0	// not used
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    //language ID = U.S. English, char set = Windows, Multilingual
    BEGIN
      VALUE "CompanyName",      "Benjamin Vernoux\0"
      VALUE "FileDescription",  "HydraSDR Software Defined Radio Library\0"
      VALUE "FileVersion",      HYDRASDR_VERSION "\0"
      VALUE "InternalName",     "hydrasdr.dll\0"
      VALUE "LegalCopyright",   "(C) 2025-2026 Benjamin Vernoux\0"
      VALUE "LegalTrademarks",  "HydraSDR\0"
      VALUE "OriginalFilename", "hydrasdr.dll\0"
      VALUE "ProductName",      "HydraSDR\0"
      VALUE "ProductVersion",   HYDRASDR_VERSION "\0"
      VALUE "Comments",         "https://www.hydrasdr.com\0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 1252
  END
END
